# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from swagger_server.models.base_model_ import Model
from swagger_server import util


class Tag(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, key: str=None, value: str=None, nanoid: str=None):  # noqa: E501
        """Tag - a model defined in Swagger

        :param key: The key of this Tag.  # noqa: E501
        :type key: str
        :param value: The value of this Tag.  # noqa: E501
        :type value: str
        :param nanoid: The nanoid of this Tag.  # noqa: E501
        :type nanoid: str
        """
        self.swagger_types = {
            'key': str,
            'value': str,
            'nanoid': str
        }

        self.attribute_map = {
            'key': 'key',
            'value': 'value',
            'nanoid': 'nanoid'
        }

        self._key = key
        self._value = value
        self._nanoid = nanoid

    @classmethod
    def from_dict(cls, dikt) -> 'Tag':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Tag of this Tag.  # noqa: E501
        :rtype: Tag
        """
        return util.deserialize_model(dikt, cls)

    @property
    def key(self) -> str:
        """Gets the key of this Tag.


        :return: The key of this Tag.
        :rtype: str
        """
        return self._key

    @key.setter
    def key(self, key: str):
        """Sets the key of this Tag.


        :param key: The key of this Tag.
        :type key: str
        """
        if key is None:
            raise ValueError("Invalid value for `key`, must not be `None`")  # noqa: E501

        self._key = key

    @property
    def value(self) -> str:
        """Gets the value of this Tag.


        :return: The value of this Tag.
        :rtype: str
        """
        return self._value

    @value.setter
    def value(self, value: str):
        """Sets the value of this Tag.


        :param value: The value of this Tag.
        :type value: str
        """
        if value is None:
            raise ValueError("Invalid value for `value`, must not be `None`")  # noqa: E501

        self._value = value

    @property
    def nanoid(self) -> str:
        """Gets the nanoid of this Tag.


        :return: The nanoid of this Tag.
        :rtype: str
        """
        return self._nanoid

    @nanoid.setter
    def nanoid(self, nanoid: str):
        """Sets the nanoid of this Tag.


        :param nanoid: The nanoid of this Tag.
        :type nanoid: str
        """
        if nanoid is None:
            raise ValueError("Invalid value for `nanoid`, must not be `None`")  # noqa: E501

        self._nanoid = nanoid



# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from swagger_server.test import BaseTestCase


class TestTagsController(BaseTestCase):
    """TagsController integration test stubs"""

    def test_tags_count_get(self):
        """Test case for tags_count_get

        Get number of tags present in MDB
        """
        response = self.client.open(
            '/v1/tags/count',
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_tags_get(self):
        """Test case for tags_get

        Get all tag nodes in MDB
        """
        response = self.client.open(
            '/v1/tags',
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from swagger_server.models.base_model_ import Model
from swagger_server import util


class Model(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, handle: str=None, name: str=None, repository: str=None, nanoid: str=None):  # noqa: E501
        """Model - a model defined in Swagger

        :param handle: The handle of this Model.  # noqa: E501
        :type handle: str
        :param name: The name of this Model.  # noqa: E501
        :type name: str
        :param repository: The repository of this Model.  # noqa: E501
        :type repository: str
        :param nanoid: The nanoid of this Model.  # noqa: E501
        :type nanoid: str
        """
        self.swagger_types = {
            'handle': str,
            'name': str,
            'repository': str,
            'nanoid': str
        }

        self.attribute_map = {
            'handle': 'handle',
            'name': 'name',
            'repository': 'repository',
            'nanoid': 'nanoid'
        }

        self._handle = handle
        self._name = name
        self._repository = repository
        self._nanoid = nanoid

    @classmethod
    def from_dict(cls, dikt) -> 'Model':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Model of this Model.  # noqa: E501
        :rtype: Model
        """
        return util.deserialize_model(dikt, cls)

    @property
    def handle(self) -> str:
        """Gets the handle of this Model.


        :return: The handle of this Model.
        :rtype: str
        """
        return self._handle

    @handle.setter
    def handle(self, handle: str):
        """Sets the handle of this Model.


        :param handle: The handle of this Model.
        :type handle: str
        """
        if handle is None:
            raise ValueError("Invalid value for `handle`, must not be `None`")  # noqa: E501

        self._handle = handle

    @property
    def name(self) -> str:
        """Gets the name of this Model.


        :return: The name of this Model.
        :rtype: str
        """
        return self._name

    @name.setter
    def name(self, name: str):
        """Sets the name of this Model.


        :param name: The name of this Model.
        :type name: str
        """

        self._name = name

    @property
    def repository(self) -> str:
        """Gets the repository of this Model.


        :return: The repository of this Model.
        :rtype: str
        """
        return self._repository

    @repository.setter
    def repository(self, repository: str):
        """Sets the repository of this Model.


        :param repository: The repository of this Model.
        :type repository: str
        """

        self._repository = repository

    @property
    def nanoid(self) -> str:
        """Gets the nanoid of this Model.


        :return: The nanoid of this Model.
        :rtype: str
        """
        return self._nanoid

    @nanoid.setter
    def nanoid(self, nanoid: str):
        """Sets the nanoid of this Model.


        :param nanoid: The nanoid of this Model.
        :type nanoid: str
        """
        if nanoid is None:
            raise ValueError("Invalid value for `nanoid`, must not be `None`")  # noqa: E501

        self._nanoid = nanoid



# coding: utf-8

from __future__ import absolute_import

from flask import json
from six import BytesIO

from swagger_server.test import BaseTestCase


class TestTagController(BaseTestCase):
    """TagController integration test stubs"""

    def test_tag_key_value_entities_count_get(self):
        """Test case for tag_key_value_entities_count_get

        Get number of entities tagged by key:value
        """
        query_string = [('skip', 56),
                        ('limit', 56)]
        response = self.client.open(
            '/v1/tag/{key}/{value}/entities/count'.format(key='key_example', value='value_example'),
            method='GET',
            query_string=query_string)
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_tag_key_value_entities_get(self):
        """Test case for tag_key_value_entities_get

        Get list of entities tagged by key:value
        """
        response = self.client.open(
            '/v1/tag/{key}/{value}/entities'.format(key='key_example', value='value_example'),
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))

    def test_tag_key_values_get(self):
        """Test case for tag_key_values_get

        Get list of tags having specified tag key
        """
        response = self.client.open(
            '/v1/tag/{key}/values'.format(key='key_example'),
            method='GET')
        self.assert200(response,
                       'Response body is : ' + response.data.decode('utf-8'))


if __name__ == '__main__':
    import unittest
    unittest.main()



# coding: utf-8

from __future__ import absolute_import
from datetime import date, datetime  # noqa: F401

from typing import List, Dict  # noqa: F401

from swagger_server.models.base_model_ import Model
from swagger_server import util


class Entity(Model):
    """NOTE: This class is auto generated by the swagger code generator program.

    Do not edit the class manually.
    """

    def __init__(self, label: str=None, nanoid: str=None):  # noqa: E501
        """Entity - a model defined in Swagger

        :param label: The label of this Entity.  # noqa: E501
        :type label: str
        :param nanoid: The nanoid of this Entity.  # noqa: E501
        :type nanoid: str
        """
        self.swagger_types = {
            'label': str,
            'nanoid': str
        }

        self.attribute_map = {
            'label': 'label',
            'nanoid': 'nanoid'
        }

        self._label = label
        self._nanoid = nanoid

    @classmethod
    def from_dict(cls, dikt) -> 'Entity':
        """Returns the dict as a model

        :param dikt: A dict.
        :type: dict
        :return: The Entity of this Entity.  # noqa: E501
        :rtype: Entity
        """
        return util.deserialize_model(dikt, cls)

    @property
    def label(self) -> str:
        """Gets the label of this Entity.


        :return: The label of this Entity.
        :rtype: str
        """
        return self._label

    @label.setter
    def label(self, label: str):
        """Sets the label of this Entity.


        :param label: The label of this Entity.
        :type label: str
        """
        if label is None:
            raise ValueError("Invalid value for `label`, must not be `None`")  # noqa: E501

        self._label = label

    @property
    def nanoid(self) -> str:
        """Gets the nanoid of this Entity.


        :return: The nanoid of this Entity.
        :rtype: str
        """
        return self._nanoid

    @nanoid.setter
    def nanoid(self, nanoid: str):
        """Sets the nanoid of this Entity.


        :param nanoid: The nanoid of this Entity.
        :type nanoid: str
        """
        if nanoid is None:
            raise ValueError("Invalid value for `nanoid`, must not be `None`")  # noqa: E501

        self._nanoid = nanoid



import connexion
from . import encoder
from importlib_resources import files

def create_app():
    app = connexion.App(__name__, specification_dir=files().joinpath('swagger'))
    app.app.json_encoder = encoder.JSONEncoder
    app.add_api(files().joinpath('swagger/swagger.yaml'), arguments={'title': 'Simple Terminology Server API'})
    return app.app  # this returns a Flask app corresponding to the Connexion app



# coding: utf-8

# flake8: noqa
from __future__ import absolute_import
# import models into model package
from swagger_server.models.entity import Entity
from swagger_server.models.model import Model
from swagger_server.models.model_property import ModelProperty
from swagger_server.models.node import Node
from swagger_server.models.tag import Tag
from swagger_server.models.term import Term



def pvs_synonyms_model_version_get(modelHandle, modelVersion):
    """
    Get Permissible Values and Synonyms for a specified model and version.

    :param modelHandle: Handle of model (use /models to find available handles)
    :type modelHandle: str
    :param modelVersion: Version of model (use /models to find available versions)
    :type modelVersion: str

    :rtype: object
    """
    return "do some magic!"


def cde_pvs_by_id_with_version_get(id, version):
    """
    Get PVs for a given CDE id and version.

    :param id: CDE ID
    :type id: str
    :param version: CDE Version
    :type version: str

    :rtype: object
    """
    return "do some magic!"


def term_by_origin_with_version_get(origin_name, origin_id, origin_version):
    """
    Get term by origin, origin_id, and origin_version.

    :param origin_name: Origin Name
    :type origin_name: str
    :param origin_id: Origin ID
    :type origin_id: str
    :param origin_version: Origin Version
    :type origin_version: str

    :rtype: object
    """
    return "do some magic!"



import sys
from flask import render_template
from .errors import bp


@bp.app_errorhandler(400)
def bad_request_error(error):
    return render_template("errors/400.html"), 400


@bp.app_errorhandler(404)
def not_found_error(error):
    return render_template("errors/404.html"), 404


@bp.app_errorhandler(500)
def internal_error(error):
    return render_template("errors/500.html"), 500



import logging
from dotenv import load_dotenv
from logging.handlers import RotatingFileHandler
from .mdb import mdb
from flask import Flask #, request, current_app
from flask_bootstrap import Bootstrap4
from flask_moment import Moment
from flask_wtf.csrf import CSRFProtect
from bento_sts.config import Config

from pdb import set_trace

bootstrap4 = Bootstrap4()
moment = Moment()
csrf = CSRFProtect()
load_dotenv(".env")
logger = logging.getLogger()


def create_app(config_class=Config):
    # see: "Application Object -> About the First Parameter"
    # at https://flask.palletsprojects.com/en/1.1.x/api/#flask.g
    app = Flask(__name__)
    app.config.from_object(config_class)

    """ or set to None for default theme """
    app.config['BOOTSTRAP_BOOTSWATCH_THEME'] = 'spacelab'

    app.config['MDB'] = mdb(
                app.config["NEO4J_MDB_URI"],
                app.config["NEO4J_MDB_USER"],
                app.config["NEO4J_MDB_PASS"],
            )

    minfo = app.config['MDB'].mdb_.get_model_info()

    app.config['MODEL_LIST'] = sorted(set([x['handle'] for x in minfo]))
    app.config['VERSIONS_BY_MODEL'] = {m: sorted(
        [v['version'] for v in minfo if v['handle'] == m])
        for m in app.config['MODEL_LIST'] }
    app.config['VERSIONS_BY_MODEL'] = {}
    app.config['LATEST_VERSION_BY_MODEL'] = {}
    # below assumes that minfo is sorted lexically by (model, version)
    for m in app.config['MODEL_LIST']:
        for info in [x for x in minfo if x['handle'] == m]:
            if app.config['VERSIONS_BY_MODEL'].get(m):
                app.config['VERSIONS_BY_MODEL'][m].append(info['version'])
            else:
                app.config['VERSIONS_BY_MODEL'][m] = [info['version']]
            if info.get('latest_version') == True or info.get('is_latest_version') == True:
                app.config['LATEST_VERSION_BY_MODEL'][m] = info['version']
    for m in app.config['MODEL_LIST']:
        if not app.config['LATEST_VERSION_BY_MODEL'].get(m):
            app.config['LATEST_VERSION_BY_MODEL'][m] = app.config['VERSIONS_BY_MODEL'][m][0]
            
    app.config['MODEL_LIST'].insert(0,"ALL")

    bootstrap4.init_app(app)
    moment.init_app(app)
    csrf.init_app(app)

    from .main import bp as main_bp
    app.register_blueprint(main_bp)
    from .api import bp as api_bp
    app.register_blueprint(api_bp)
    from .errors import bp as errors_bp
    app.register_blueprint(errors_bp)

    if not app.debug and not app.testing:
        file_handler = RotatingFileHandler(
            app.config['FLASK_LOGFILE'], maxBytes=10240, backupCount=10
        )
        file_handler.setFormatter(
            logging.Formatter(
                "%(asctime)s %(levelname)s: %(message)s"
                " [in %(pathname)s:%(lineno)d]"
            )
        )
        file_handler.setLevel(logging.INFO)
        app.logger.addHandler(file_handler)

        app.logger.setLevel(logging.DEBUG)
        app.logger.info("STS startup")

    return app



import connexion
import six

from swagger_server.models.model import Model  # noqa: E501
from swagger_server import util


def models_count_get():  # noqa: E501
    """Get number of available models

     # noqa: E501


    :rtype: object
    """
    return 'do some magic!'


def models_get():  # noqa: E501
    """Get info on available models

     # noqa: E501


    :rtype: List[Model]
    """
    return 'do some magic!'



import connexion
import six

from swagger_server.models.entity import Entity  # noqa: E501
from swagger_server import util


def id_id_get(id):  # noqa: E501
    """Get MDB entity with specified nanoid

     # noqa: E501

    :param id: Nanoid (6 character unique string)
    :type id: str

    :rtype: Entity
    """
    return 'do some magic!'



import subprocess


def get_yaml_for(model):

    result = subprocess.run(["perl", "-X", "../Bento-STS-CLI/bin/yaml.pl", '-model', model], stdout=subprocess.PIPE)
    return result.stdout.decode('utf-8')



"""
makeq - make a Neo4j query from an endpoint path or JSON payload.
"""
import yaml

from pdb import set_trace
from ._engine import _engine

class Query(object):
    paths = {}
    cache = {}

    def __init__(self, qspec):
        if isinstance(qspec, str):
            path = qspec
            if path.startswith("/"):
                path = path[1:]
            self.toks = path.split("/")
            self._engine = None
            if not self._engine:
                self._engine = _engine()
                if not self._engine.parse(self.toks):
                    raise RuntimeError(self._engine.error)
        elif isinstance(qspec, dict):
            pass
        else:
            raise RuntimeError("Query arg1 must be path string or dict")

    @classmethod
    def set_paths(cls, paths):
        if paths.get('paths'):
            cls.paths = paths['paths']
        else:
            cls.paths = paths
        _engine.set_paths(cls.paths)
        return True

    @classmethod
    def load_paths(cls, flo):
        p = yaml.load(flo, Loader=yaml.CLoader)
        return cls.set_paths(p)


    @property
    def statement(self):
        return self._engine.statement

    @property
    def params(self):
        return self._engine.params

    @property
    def path_id(self):
        return self._engine.path_id
    
    def __str__(self):
        return str(self.statement)


def f(pfx, pth):
    tok = [x for x in pth if x.startswith('$')]
    if not tok:
        tok = [x for x in pth if not x.startswith('_')]
    if not tok:
        print(pfx)
        return
    else:
        if pth.get('_return'):
            print(pfx)
        for t in tok:
            f('/'.join([pfx, t]), pth[t])
        return



from __future__ import annotations

import functools

from bento_meta.mdb import SearchableMDB
from bento_meta.model import Model
from flask import current_app, url_for

plural = {
    "node": "nodes",
    "edge": "edges",
    "relationship": "relationships",
    "property": "properties",
    "tag": "tags",
    "origin": "origins",
    "concept": "concepts",
    "predicate": "predicates",
    "value_set": "value_sets",
    "term": "terms",
}


class mdb:
    """
    Read functionality for driving STS UI. Mixins mdb_update and mdb_tags
    could be used here for write and tag functionality.
    """

    mdb_ = None
    term_values = None

    def __init__(self, uri, user, pw):
        if mdb.mdb_ is None:
            mdb.mdb_ = SearchableMDB(uri, user, pw)
        self.mdb = mdb.mdb_

    def close(self):
        self.mdb.close()

    def get_list_of_models(self):
        models = self.mdb.get_model_nodes()
        ret = {}
        if models:
            for mdl in models:
                h = mdl["m"]["handle"]
                if ret.get(h):
                    ret[h]["version"].append(mdl["m"]["version"])
                else:
                    ret[h] = mdl["m"]
                    if not ret[h].get("name"):
                        ret[h]["name"] = ret[h]["handle"]
                    ret[h]["version"] = [mdl["m"]["version"]]
        return [x for x in ret.values()]

    def get_model_by_name(self, name):
        model_nodes = self.mdb.get_model_nodes(model=name)
        models = []
        for node in model_nodes:
            model = Model(handle=name)
            model.version = node["m"].get("version")
            model.repository = node["m"].get("repository")
            models.append(model)
        return models

    # ####################################################################### #
    # NODES
    # ####################################################################### #
    def get_node_by_id(self, nid, model=None):
        props_result = self.mdb.get_node_and_props_by_node_id(nid)
        edges_result = self.mdb.get_node_edges_by_node_id(nid)
        to_nodes = {}
        from_nodes = {}
        if not props_result and not edges_result:
            return {}
        result = {
            "id": nid,
            "handle": props_result[0]["handle"],
            "model": props_result[0]["model"],
            "version": props_result[0].get("version"),
            "has_properties": [],
            "has_relationship_to_nodes": [],
            "has_relationship_from_nodes": [],
        }
        if props_result:
            for p in props_result[0]["props"]:
                result["has_properties"].append(
                    {
                        "id": p["nanoid"],
                        "handle": p["handle"],
                        "model": p["model"],
                        "version": p.get("version"),
                        "type": "property",
                        "link": url_for(
                            "main.entities",
                            entities="properties",
                            id=p["nanoid"],
                        ),
                        "value_domain": p["value_domain"],
                    },
                )
        if edges_result:
            for e in edges_result:
                node = e["far_node"]
                if node:
                    rln = e["rln"]
                    add_to = None
                    if e["far_type"] == "has_src":
                        add_to = from_nodes
                    elif e["far_type"] == "has_dst":
                        add_to = to_nodes
                    else:
                        RuntimeError("What?")
                    add_to[node["nanoid"]] = {
                        "id": node["nanoid"],
                        "handle": node["handle"],
                        "model": node["model"],
                        "version": node.get("version"),
                        "link": url_for(
                            "main.entities",
                            entities="nodes",
                            id=node["nanoid"],
                        ),
                        "type": "node",
                        "relationship": {
                            "id": rln["nanoid"],
                            "handle": rln["handle"],
                            "model": rln["model"],
                            "version": rln.get("version"),
                            "link": "/relationships/{}".format(rln["nanoid"]),
                            "type": "relationship",
                        },
                    }
            result["has_relationship_to_nodes"].extend(to_nodes.values())
            result["has_relationship_from_nodes"].extend(from_nodes.values())
        return result

    # ----------------------------------------------------------------------- #

    def get_list_of_nodes(self, model=None, version=None):
        """
        In [3]: m.get_list_of_nodes()
        Out[3]:
        [{'yXWr0Y': 'study_site'},
         {'N0Qx7Z': 'off_study'},
         {'nUoHJH': 'diagnosis
        """
        if model == "ALL":
            model = None
            version = None
        elif version == "ALL":
            version = "*"
        result = self.mdb.get_nodes_by_model(model, version)
        if result:
            return [
                (x["nanoid"], x["handle"], x["model"], x["version"]) for x in result
            ]
        return []

    # ####################################################################### #
    # PROPERTIES
    # ####################################################################### #

    def get_list_of_properties(self, model=None, version=None):
        if model == "ALL":
            model = None
            version = None
        elif version == "ALL":
            version = "*"
        np_result = self.mdb.get_nodes_and_props_by_model(model, version)
        if not np_result:
            return []
        result = []
        for np in np_result:
            result.extend(
                [
                    {
                        "prop_id": p["nanoid"],
                        "prop_handle": p["handle"],
                        "node_model": np["model"],
                        "node_handle": np["handle"],
                        "node_id": np["id"],
                        "node_version": np.get("version"),
                    }
                    for p in np["props"]
                ],
            )
        return result

    def get_property_by_id(self, pid, model=None):
        p_result = self.mdb.get_with_statement(
            "match (p:property {nanoid:$pid})<-[:has_property]-(n:node) "
            "with p,n "
            "optional match (p)-[:has_concept]->(:concept)<-[:represents]-(a:term) "
            "optional match (p)-[:has_value_set]->(vs:value_set)-[:has_term]->(t:term) "
            "return p.nanoid as id, p.handle as handle, p.model as model, "
            "p.version as version, "
            "p.value_domain as value_domain, p as prop, n as node, "
            "  vs as value_set, collect(distinct a) as annots, collect(t) as terms",
            {"pid": pid}
            )
        if not p_result:
            return {}
        pr = p_result[0]
        pr["type"] = "property"
        pr["link"] = "/properties/{}".format(pr["id"])
        # pr["version"] = pr.get("version"),
        if pr["value_set"]:
            pr["has_valueset"] = {
                "id": pr["value_set"]["nanoid"],
                "type": "valueset",
                "link": url_for(
                    "main.entities",
                    entities="valuesets",
                    id=pr["value_set"]["nanoid"],
                ),
            }
        if pr["terms"]:
            pr["has_terms"] = [{"id": t["nanoid"], "value": t["value"],
                                    "origin": t["origin_name"] if "origin_name" in t else None,
                                    "type": "term",
                                    "link": url_for("main.entities",
                                                    entities='terms',
                                                    id=t["nanoid"])}
                               for t in pr["terms"] if 'nanoid' in t]
        if pr["annots"]:
            pr["has_annots"] = [{"id": t["nanoid"], "value": t["value"],
                                    "origin": t["origin_name"] if "origin_name" in t else None,
                                    "type": "term",
                                    "link": url_for("main.entities",
                                                    entities='terms',
                                                    id=t["nanoid"])}
                               for t in pr["annots"] if 'nanoid' in t]
        return pr

    # ####################################################################### #
    # VALUESETS
    # ####################################################################### #

    def get_valueset_by_id(self, vsid, model=None):
        vs_result = self.mdb.get_valueset_by_id(vsid)
        if not vs_result:
            return {}
        result = {
            "id": vs_result[0]["nanoid"],
            "handle": vs_result[0]["handle"],
            "model": model,  # kludge
            "url": vs_result[0]["url"],
            "desc": vs_result[0]["desc"],
            "type": "valueset",
            "link": url_for(
                "main.entities",
                entities="valuesets",
                id=vs_result[0]["nanoid"],
            ),
            "_for_propertyhandle": "_for_propertyhandle",  # kludge
            "_for_propertyid": "_for_propertyid",  # kludge
        }
        result["has_property"] = [
            {
                "id": x["nanoid"],
                "handle": x["handle"],
                "type": "property",
                "model": x["model"],
                "link": "/properties/{}".format(x["nanoid"]),
            }
            for x in vs_result[0].props
        ]
        result["has_terms"] = [
            {
                "id": x["nanoid"],
                "value": x["value"],
                "origin": x["origin_name"] if "origin_name" in x else None,
                "type": "term",
                "link": url_for("main.entities", entities="terms", id=x["nanoid"]),
            }
            for x in vs_result[0].terms
        ]
        if model:
            result["has_property"] = [
                x for x in result["has_property"] if x["model"] == model
            ]
        # kludge - original assumption is that a valueset has only one
        # associated property, but this is not necessarily true.
        if result["has_property"]:
            result["_for_propertyhandle"] = result["has_property"][0]["handle"]
            result["_for_propertyid"] = result["has_property"][0]["nanoid"]
        return result

    # ----------------------------------------------------------------------- #

    def get_list_of_valuesets(self, model=None, version=None):
        vs_result = self.mdb.get_valuesets_by_model(model, version)
        if not vs_result:
            return []
        # kludge assuming that there is only one property assoc with each
        # valueset:
        return [
            {"id": x["value_set"]["nanoid"], "handle": x["props"][0]["handle"]}
            for x in vs_result
        ]

    # ####################################################################### #
    # TERMS
    # ####################################################################### #

    def get_term_batch_info(self, nbatches, start=0, ct=None):
        batches = []
        tabnames = []
        if not mdb.term_values:
            try:
                res = self.mdb.get_with_statement(
                    "match (t:term) with t.value as val return val order by val",
                    {},
                )
            except Exception as e:
                raise e
            mdb.term_values = [x["val"] for x in res]
        if ct is None:
            ct = len(self.term_values)
        bsize = ct // nbatches or ct
        for n in range(start, min(start + ct, len(self.term_values)), bsize):
            l = min(n + bsize - 1, len(self.term_values) - 1)
            batches.append(
                {
                    "first": n,
                    "first_term": self.term_values[n],
                    "last": l,
                    "last_term": self.term_values[l],
                },
            )

        # if n+bsize != start+ct:
        #     batches[-1]['last'] = start+ct-1
        #     batches[-1]['last_term'] = self.term_values[start+ct-1]

        for n in range(len(batches)):
            v1 = batches[n]["first_term"] or "___"
            v2 = batches[n]["last_term"] or "___"
            tabnames.append(f"{v1[0:3]}-{v2[0:3]}")

        return (batches, tabnames)

    @functools.lru_cache
    @staticmethod
    def get_term_batch(start, bsize):
        try:
            res = mdb.mdb_.get_with_statement(
                "match (t:term) with t order by t.value "
                "return t as term skip $start limit $bsize",
                {"start": start, "bsize": bsize},
            )
        except Exception as e:
            raise e
        return res

    def get_term_by_id(self, tid):
        t_result = self.mdb.get_with_statement(
            "match (t:term {nanoid:$id}) "
            "optional match (t)<-[:has_term]-(v:value_set)<-[:has_value_set]-(p:property) "
            "with t as term, t.origin_name as oname,  collect(p) as props "
            "optional match (o:origin) where o.name = oname "
            "return term, o as origin, props ",
            {"id": tid},
        )
        if not t_result:
            return {}
        result = t_result[0]
        result["props"] = sorted(
            result["props"],
            key=lambda x: (x["handle"], x["model"], x.get("version")),
        )
        s_result = self.mdb.get_with_statement(
            "match (t:term {nanoid:$id})-[:represents]->(c:concept)<-[:represents]-(s:term) "
            "optional match (g:tag {key:'mapping_source'})<-[:has_tag]-(c) "
            "return s.value as value, s.nanoid as id, s.origin_name as origin, collect(g.value) as sources",
            {"id": tid},
        )
        if s_result:
            result["synonyms"] = sorted(s_result, key=lambda x: x["value"])
        else:
            result["synonyms"] = None
        return result

    def get_list_of_terms(self, start=None, end=None):
        t_result = self.mdb.get_with_statement(
            "match (t:term)<-[:has_term]-(:value_set) "
            "with distinct t as t "
            "return t.value as value, t as term",
            {},
        )
        return t_result

    # ####################################################################### #
    # ORIGINS
    # ####################################################################### #

    def get_list_of_origins(self, dummy=None):
        origins = self.mdb.get_origins()
        result = [{x["o"]["nanoid"]: x["o"]["name"]} for x in origins]
        if result:
            return result
        return []

    def get_origin_by_id(self, oid):
        result = self.mdb.get_origin_by_id(oid)
        return result

    # TAGS

    def get_tagged_entities(self, tag_key, tag_value=None):
        props = "{key:$key, value:$value}" if tag_value else "{key:$key}"
        ents_by_tag = self.mdb.get_with_statement(
            f"match (t:tag {props}) "
            "with t "
            "match (e)-[:has_tag]->(t) "
            "with t, {type:head(labels(e)), ent:e} as ee "
            "return t.key as tag_key, t.value as tag_value, collect(ee) as entities",
            {"key": tag_key, "value": tag_value},
        )

        for i in range(len(ents_by_tag)):
            for j in range(len(ents_by_tag[i]["entities"])):
                ents_by_tag[i]["entities"][j]["plural"] = plural[
                    ents_by_tag[i]["entities"][j]["type"]
                ]
        return ents_by_tag

    def get_tags_and_values(self):
        tags = self.mdb.get_tags_and_values()
        return tags

    # SEARCH

    def search_entity_handles(self, qstring):
        res = {}
        res["nodes"] = self.mdb.query_index("nodeHandle", qstring) or []
        res["properties"] = self.mdb.query_index("propHandle", qstring) or []
        res["relationships"] = self.mdb.query_index("edgeHandle", qstring) or []
        return res

    def search_terms(self, qstring, search_values=True, search_definitions=True):
        return self.mdb.search_terms(
            qstring,
            search_values=search_values,
            search_definitions=search_definitions,
        )

    # ####################################################################### #
    # CDE PVs & Synonyms
    # ####################################################################### #

    @functools.lru_cache
    @staticmethod
    def get_model_pvs_synonyms(model: str | None = None, version: str | None = None):
        qry = (
            "MATCH (n {model: $dataCommons, version: $version})-[:has_property]->"
            "(p:property) "
            "WITH collect(p) AS props "
            "UNWIND props AS prop "
            "OPTIONAL MATCH (prop)-[:has_concept]->(c:concept)<-[:represents]-"
            "(cde:term) WHERE toLower(cde.origin_name) CONTAINS 'cadsr' "
            "OPTIONAL MATCH (prop)-[:has_value_set]->(:value_set)-[:has_term]->(t:term)"
            " WITH prop, cde.origin_id AS CDECode, cde.origin_version AS CDEVersion, "
            "cde.value AS CDEFullName, cde.origin_id + '|' + "
            "COALESCE(cde.origin_version, '') AS cde_hdl, collect(t) AS model_pvs, "
            "CASE WHEN cde IS NOT NULL THEN true ELSE false END AS has_cde "
            "OPTIONAL MATCH (v:value_set {handle: cde_hdl})-[:has_term]->(cde_pv:term) "
            "WITH prop, CDECode, CDEVersion, CDEFullName, model_pvs, has_cde, "
            "collect(cde_pv) AS cde_pvs WITH prop, CDECode, CDEVersion, CDEFullName, "
            "model_pvs, has_cde, cde_pvs, CASE WHEN has_cde AND size(cde_pvs) > 0 "
            "AND NONE(p in cde_pvs WHERE p.value =~ 'https?://.*') THEN cde_pvs "
            "WHEN has_cde and size(cde_pvs) > 0 "
            "AND ANY(p in cde_pvs WHERE p.value =~ 'https?://.*') "
            "AND size(model_pvs) > 0 THEN model_pvs "
            "WHEN NOT has_cde AND size(model_pvs) > 0 THEN model_pvs "
            "ELSE [null] END AS pvs "
            "UNWIND pvs AS pv "
            "OPTIONAL MATCH (pv)-[:represents]->(c:concept)<-[:represents]-"
            "(syn:term), (c)-[:has_tag]->(g:tag {key: 'mapping_source', value: 'NCIm'})"
            " WHERE pv <> syn and pv.value <> syn.value "
            "WITH prop, CDECode, CDEVersion, CDEFullName, model_pvs, "
            "pv.value AS pv_val, collect(DISTINCT syn.value) AS syn_vals "
            "WITH prop, CDECode, CDEVersion, CDEFullName, model_pvs, "
            "collect({value: pv_val, synonyms: syn_vals}) AS formatted_pvs "
            "RETURN $dataCommons AS dataCommons, $version AS version, "
            "prop AS property, CDECode, CDEVersion, CDEFullName, "
            "formatted_pvs AS permissibleValues"
        )

        parms = {"dataCommons": model, "version": version}

        return mdb.mdb_.get_with_statement(qry, parms)

    @functools.lru_cache
    @staticmethod
    def get_cde_pvs_by_id(id: str | None = None, version: str | None = None):
        """Get CDE PVs for a given CDE id and optional version."""
        qry = (
            "MATCH (cde:term {origin_id: $cde_id}) "
            "WHERE ($cde_version = '' OR cde.origin_version = $cde_version) "
            "OPTIONAL MATCH (vs:value_set)-[:has_term]->(pv:term) "
            "WHERE vs.handle = $cde_id + '|' + coalesce(cde.origin_version, '') "
            "RETURN cde, vs.url as value_set_url, COLLECT(pv) as pvs"
        )
        parms = {"cde_id": id, "cde_version": version}

        return mdb.mdb_.get_with_statement(qry, parms)

    @functools.lru_cache
    @staticmethod
    def get_term_nanoid_by_origin(
        origin_name: str | None = None,
        origin_id: str | None = None,
        origin_version: str | None = None,
    ):
        """Get terms by origin, origin_id, and origin_version."""
        qry = (
            "MATCH (t:term) WHERE t.origin_name = $origin_name "
            "AND t.origin_id = $origin_id AND ($origin_version IS NULL "
            "OR $origin_version = '' OR t.origin_version = $origin_version) "
            "WITH t ORDER by split(t.origin_version, '.') DESC "
            "RETURN t.nanoid as term_nanoid"
        )
        parms = {
            "origin_name": origin_name,
            "origin_id": origin_id,
            "origin_version": origin_version,
        }

        return mdb.mdb_.get_with_statement(qry, parms)



from flask import request
from flask_wtf import FlaskForm
from wtforms import SelectField, StringField, SubmitField
from wtforms.validators import DataRequired


class SearchForm(FlaskForm):
    qstring = StringField("Query", validators=[DataRequired()])
    terms = SubmitField("Search Terms")
    models = SubmitField("Search Models")

    def __init__(self, *args, **kwargs):
        if "formdata" not in kwargs:
            kwargs["formdata"] = request.args
        if "csrf_enabled" not in kwargs:
            kwargs["csrf_enabled"] = False
        super(SearchForm, self).__init__(*args, **kwargs)


class SelectModelForm(FlaskForm):
    model = SelectField("Model")
    version = SelectField("Version")
    filter = SubmitField("Filter")
    export = SubmitField("Export JSON")


class SelectVersionForm(FlaskForm):
    version = SelectField('Version')
    submit_version = SubmitField('Version')
    



        



import re

from flask import (
    abort,
    current_app,
    g,
    jsonify,
    redirect,
    render_template,
    request,
    url_for,
)
from flask_paginate import Pagination, get_page_parameter

from . import bp
from .forms import (
    SearchForm,
    SelectModelForm,
    SelectVersionForm,
)

def mdb():
    # only called in app context
    return current_app.config['MDB']

@bp.before_app_request
def before_request():
    g.search_form = SearchForm()
    g.locale = "EN_US"


@bp.route("/", methods=["GET", "POST"])
@bp.route("/index", methods=["GET", "POST"])
def index():
    return render_template(
        "index.html",
        title="Home",
    )


# - - - - - - - - - - - - - - - - - - - - - - - - - - - - -


@bp.route("/<entities>", defaults={'id': None}, methods=['GET','POST'])
@bp.route('/<entities>/<id>', methods=['GET','POST'])

@bp.route("/<entities>", defaults={"id": None}, methods=["GET", "POST"])
@bp.route("/<entities>/<id>", methods=["GET", "POST"])
def entities(entities, id):
    format = request.args.get("format") or request.form.get("format") or ""
    if request.form.get("export"):
        format = "json"

    model = request.args.get("model") or request.form.get("model") or "ALL"
    version = request.args.get("version") or request.form.get("version") or None
    id_ = request.args.get("id")
    page = request.args.get(get_page_parameter(), type=int, default=1)
    select_form = SelectModelForm()
    select_form.model.choices = [(x, x) for x in current_app.config["MODEL_LIST"]]
    if model != "ALL":
        select_form.version.choices = [
            (x, x) for x in current_app.config["VERSIONS_BY_MODEL"][model]
        ]
        select_form.version.choices.insert(0, ("ALL", "ALL"))
    else:
        select_form.version.choices = [("ALL", "ALL")]
        version = "*"

    current_app.logger.info(
        f"model: {model}, version: {version}, page: {page}, format: {format}",
    )
    id = id or id_
    dispatch = {
        "nodes": {
            "title": "Node",
            "list_title": "Nodes",
            "template": "mdb-node.html",
            "subtype": "nodes",
            "sort_key": lambda x: (x[1], x[2], x[3]),
            "get_by_id": mdb().get_node_by_id,
            "get_list": mdb().get_list_of_nodes,
        },
        "properties": {
            "title": "Property",
            "list_title": "Properties",
            "template": "mdb-property.html",
            "subtype": "properties",
            "sort_key": lambda x: (x["prop_handle"], x["node_model"], x["node_handle"]),
            "display": "prop-tuple",
            "get_by_id": mdb().get_property_by_id,
            "get_list": mdb().get_list_of_properties,
        },
        "valuesets": {
            "title": "Value Set",
            "list_title": "Value Sets",
            "template": "mdb-valueset.html",
            "subtype": "valuesets",
            "sort_key": lambda x: x["handle"],
            "display": "list",
            "get_by_id": mdb().get_valueset_by_id,
            "get_list": mdb().get_list_of_valuesets,
        },
        "terms": {
            "title": "Term",
            "list_title": "Terms",
            "template": "mdb-term.html",
            "subtype": "terms",
            "sort_key": lambda x: (
                x["value"].lower() if type(x["value"]) == str else ""
            ),
            "display": "term-tuple",
            "get_by_id": mdb().get_term_by_id,
            "get_list": mdb().get_list_of_terms,
        },
        "origins": {
            "title": "Origin",
            "list_title": "Origins",
            "template": "mdb-origin.html",
            "subtype": "origins",
            "sort_key": lambda x: list(x.values())[0],
            "display": "list",
            "get_list": mdb().get_list_of_origins,
            "get_by_id": mdb().get_origin_by_id,
        },
    }

    # A: single entity
    if id is not None:
        ent_ = dispatch[entities]["get_by_id"](id)
        if ent_ is None or not bool(ent_):
            return render_template("/errors/400.html"), 400
        if type(ent_) == dict and ent_.get("has_properties"):  # nodes only kludge
            ent_["has_properties"].sort(key=lambda x: x["handle"])
        if format == "json":
            return jsonify(ent_)
        return render_template(
            dispatch[entities]["template"],
            title=dispatch[entities]["title"],
            mdb=ent_,
            subtype=dispatch[entities]["subtype"],
            display="detail",
        )

    # B: list, filter by model
    get_list_args = (model, version)
    if entities in ["origins", "terms"]:
        get_list_args = ()
    ents_ = dispatch[entities]["get_list"](*get_list_args)

    if format == "json":
        return jsonify(ents_)
    pgurl = url_for("main.entities", entities=entities)
    pgurl += "?page={0}"
    if model:
        pgurl += f"&model={model}"
    pagination = Pagination(
        page=page,
        total=len(ents_),
        record_name=entities,
        href=pgurl,
    )
    rendered = render_template(
        "mdb.html",
        title=dispatch[entities]["list_title"],
        mdb=sorted(ents_, key=dispatch[entities]["sort_key"]),
        subtype=dispatch[entities]["subtype"],
        display=dispatch[entities].get("display") or "tuple",
        first=(pagination.page - 1) * pagination.per_page,
        last=min((pagination.page) * pagination.per_page, len(ents_)),
        pagination=pagination,
        form=select_form,
        model=model,
    )
    # get a load of THIS kludge, dude.
    if model:
        safe_model = re.escape(model)
        rendered = re.sub(
            f'option value="{safe_model}"',
            f'option selected="true" value={model}',
            rendered,
        )

    return rendered


# ---------------------------------------------------------------------------


@bp.route("/models", methods=['GET'])
@bp.route("/models/<name>", methods=['GET','POST'])
def models(name=None):
    m = mdb()
    if name is not None:
        if name not in current_app.config["MODEL_LIST"]:
            return render_template('/errors/400.html'), 400
        version = request.args.get("version") or request.form.get("version") or current_app.config["LATEST_VERSION_BY_MODEL"][name]
        select_form = SelectVersionForm()
        select_form.version.choices = current_app.config["VERSIONS_BY_MODEL"][name]
        return render_template(
            "mdb-model.html",
            title="Model: {}".format(name),
            name=name,
            version=version,
            mdb={"nodes":m.get_list_of_nodes(name, version),
                 "props":m.get_list_of_properties(name, version)},
            subtype="main.models",
            display="detail",
            form=select_form,
        )

    else:
        models_ = m.get_list_of_models()
        return render_template(
            "mdb-model.html",
            title="Models",
            mdb=sorted(models_, key=lambda x:x["name"]),
            subtype="main.models",
            display="list",
        )


@bp.route("/terms", defaults={'start': 0}, methods=['GET', 'POST'])
@bp.route("/terms/batch/<start>", methods=['GET', 'POST'])
def terms(start, num=15):
    page = request.args.get(get_page_parameter(), type=int, default=1)
    (batches, tabnames) = mdb().get_term_batch_info(num)
    activetab = -1
    activesubtab = -1
    subbatches = None
    subtabnames = None
    batch = None
    paging = {}
    if start is not None:
        start = int(start)
        bsize = batches[0]["last"] - batches[0]["first"] + 1
        for i in range(len(batches)):
            if batches[i]["first"] <= start <= batches[i]["last"]:
                activetab = i
                break
        (subbatches, subtabnames) = mdb().get_term_batch_info(
            num,
            batches[i]["first"],
            bsize,
        )
        for j in range(len(subbatches)):
            if subbatches[j]["first"] <= start < subbatches[j]["last"]:
                activesubtab = j
                break
        sbsize = subbatches[j]["last"] - subbatches[j]["first"] + 1
        batch = type(mdb()).get_term_batch(start, sbsize)
        pgurl = url_for("main.terms", start=start)
        pgurl += "?page={0}"
        pagination = Pagination(
            page=page,
            total=len(batch),
            record_name=entities,
            href=pgurl,
        )
        paging["pagination"] = pagination
        paging["first"] = (pagination.page - 1) * pagination.per_page
        paging["last"] = min((pagination.page) * pagination.per_page, len(batch))

    return render_template(
        "mdb-term-tabs.html",
        title="Terms",
        tabnames=tabnames,
        subtabnames=subtabnames,
        batches=batches,
        subbatches=subbatches,
        activetab=activetab,
        activesubtab=activesubtab,
        batch=batch,
        paging=paging,
    )


@bp.route("/tags", defaults={"key": None, "value": None}, methods=["GET", "POST"])
@bp.route("/tags/<key>", methods=["GET", "POST"], defaults={"value": None})
@bp.route("/tags/<key>/<value>", methods=["GET", "POST"])
def tags(key=None, value=None, model=None):
    key = key or request.args.get("key")
    val = value or request.args.get("value")
    model = model or request.args.get("model")
    ents = []
    format = request.args.get("format")
    if request.form.get("format"):
        format = request.args.get("format")
    elif request.form.get("export"):
        format = "json"
    else:
        pass

    if key:
        ents = mdb().get_tagged_entities(key, val)
    else:
        ents = mdb().get_tags_and_values()

    if format == "json":
        return jsonify(ents)
    if key:
        return render_template(
            "mdb-tag.html",
            title="Tagged Entities",
            key=key,
            value=val,
            ents=ents,
            display="entities",
        )
    return render_template("mdb-tag.html", title="Tags", ents=ents, display="tags")


@bp.route("/search")
def search():
    if not g.search_form.validate():
        return redirect(url_for("main.index"))
    format = request.args.get("format")
    if request.form.get("format"):
        format = request.args.get("format")
    elif request.form.get("export"):
        format = "json"
    else:
        pass

    qstring = request.args.get("qstring")
    entdisplay = request.args.get("entdisplay") or "nodes"
    ents = []
    thing = ""
    if request.args.get("terms"):
        ents = mdb().search_terms(qstring)
        thing = "terms"
    elif request.args.get("models"):
        ents = mdb().search_entity_handles(qstring)
        thing = "models"
    else:
        abort(400)
    if format == "json":
        return jsonify(ents)
    paging = {}
    pg_tot = 0
    activetab = 'nodes'

    if not ents:
        thing = "no_hits"
    if thing == "terms":
        pg_tot = len(ents)
        pagination = Pagination(
            page=request.args.get("page", 1, type=int),
            record_name="Hits",
            total=pg_tot,
            per_page=current_app.config["HITS_PER_PAGE"],
        )
        paging["terms"] = {"pagination": pagination}
        paging["terms"]["first"] = (pagination.page - 1) * pagination.per_page
        paging["terms"]["last"] = min(pagination.page * pagination.per_page, len(ents))
        entdisplay = "terms"
    elif thing == "models":
        paging = None
        if len(ents['nodes']) == 0:
            if len(ents['properties']) > 0:
                activetab = 'properties'
            elif len(ents['relationships']) > 0:
                activetab = 'relationships'
            else:
                thing = "no_hits"
        # for ent in ("nodes", "properties", "relationships"):
        #     pagination = Pagination(
        #         page=request.args.get("page",1,type=int),
        #         record_name="Hits",
        #         total=len(ents[ent]),
        #         anchor=ent,
        #         per_page=current_app.config["HITS_PER_PAGE"],
        #     )
        #     paging[ent] = {"pagination": pagination}
        #     paging[ent]['first'] = (pagination.page-1)*pagination.per_page
        #     paging[ent]['last'] = min(pagination.page*pagination.per_page,len(ents[ent]))

    return render_template(
        "search.html",
        title="Search",
        ents=ents,
        npr=['nodes','properties','relationships'],
        thing=thing,
        q=qstring,
        activetab=activetab,
        entdisplay=entdisplay,
        paging=paging,
    )


@bp.route("/about-mdb")
def about_mdb():
    return render_template("about-mdb.html", title="About MDB")


@bp.route("/about-sts")
def about_sts():
    return render_template("about-sts.html", title="About STS")


@bp.errorhandler(413)
def too_large(e):
    return "File is too large", 413


@bp.route("/reports", methods=["GET", "POST"])
def reports():
    return render_template("reports.html")


@bp.route("/versionhistory", methods=["GET", "POST"])
def versionhistory():
    return render_template("version-history.html")


@bp.route(
    "/model-pvs/<model>/<version>",
    methods=["GET", "POST"],
    strict_slashes=False,
)
def cde_pvs_and_synonyms_by_model(model, version):
    """
    Get PVs and synonyms for a given model and version.

    Follows Data Hub logic for using PVs from CDE or model.
    """
    model = model or request.args.get("model") or request.form.get("model") or "ALL"
    version = (
        version or request.args.get("version") or request.form.get("version") or None
    )

    ents = []
#    mdb()  # instantiate so mdb.mdb_ is available for get_cde_pvs

    fmt = request.args.get("format")
    if request.form.get("format"):
        fmt = request.args.get("format")
    elif request.form.get("export"):
        fmt = "json"
    else:
        pass

    ents = mdb().get_model_pvs_synonyms(model, version)

    if fmt == "json":
        # remove attrs other than values from props
        return jsonify(
            [{**item, "property": item["property"].get("handle", "")} for item in ents],
        )
    return render_template(
        "mdb-model-pvs.html",
        title="CDE Permissible Values and Synonyms by Model",
        ents=ents,
        display="cdes",
        model=model,
        version=version,
    )


@bp.route("/cde-pvs/<id>/<version>", methods=["GET", "POST"], strict_slashes=False)
@bp.route(
    "/cde-pvs/<id>",
    defaults={"version": None},
    methods=["GET", "POST"],
    strict_slashes=False,
)
def cde_pvs_by_id(id, version):
    """Get PVs for a given CDE id and optional version."""
    cde_id = id or request.args.get("id")
    cde_version = version or request.args.get("version") or ""
    ents = []
#    mdb()  # instantiate so mdb.mdb_ is available for get_cde_pvs

    fmt = request.args.get("format")
    if request.form.get("format"):
        fmt = request.args.get("format")
    elif request.form.get("export"):
        fmt = "json"
    else:
        pass

    ents = mdb().get_cde_pvs_by_id(cde_id, cde_version)

    if fmt == "json":
        return jsonify(
            [
                {
                    "CDECode": item["cde"].get("origin_id", ""),
                    "CDEVersion": item["cde"].get("origin_version", ""),
                    "CDEFullName": item["cde"].get("value", ""),
                    "permissibleValues": [pv.get("value", "") for pv in item["pvs"]],
                }
                for item in ents
            ],
        )
    return render_template(
        "mdb-cde-pvs.html",
        title="Common Data Element (CDE) Permissible Values",
        ents=ents,
        display="cde_pvs",
        id=cde_id,
        version=cde_version,
    )


@bp.route(
    "/term-by-origin/<origin_name>/<origin_id>",
    defaults={"origin_version": None},
    methods=["GET", "POST"],
    strict_slashes=False,
)
@bp.route(
    "/term-by-origin/<origin_name>/<origin_id>/<origin_version>",
    methods=["GET", "POST"],
    strict_slashes=False,
)
def term_by_origin(origin_name, origin_id, origin_version):
    """Get term by origin, origin_id, and origin_version."""
    origin_name = origin_name or request.args.get("origin_name")
    origin_id = origin_id or request.args.get("origin_id")
    origin_version = origin_version or request.args.get("origin_version") or ""

 #   mdb()  # instantiate so mdb.mdb_ is available for get_cde_pvs

    term_nanoids = (
        mdb().get_term_nanoid_by_origin(origin_name, origin_id, origin_version) or []
    )
    if term_nanoids:
        term_nanoid = term_nanoids[0]["term_nanoid"]
    if len(term_nanoids) > 1:
        msg = (
            f"More than one Term found for origin: {origin_name} "
            f"with id: {origin_id} and version: {origin_version}. "
            "Displaying first result only."
        )
        current_app.logger.warning(msg)

    return entities(entities="terms", id=term_nanoid)
